@@ -1,35 +1,12 @@
 """Abstract base models used by the page management application."""
 from django.db import models
 from django.shortcuts import render
-from django.utils.crypto import constant_time_compare, salted_hmac
 from watson.search import SearchAdapter
 
 from cms.apps.media.models import ImageRefField
 from cms.models.managers import (OnlineBaseManager, PageBaseManager,
                                  PublishedBaseManager, SearchMetaBaseManager)
-
-
-class PathTokenGenerator:
-    '''
-        A simple token generator that takes a path and generates a hash for it.
-        Intended for use by the CMS publication middleware and OnlineBase derivatives.
-        In reality it just takes a string so it can be used for other purposes.
-    '''
-    key_salt = "django.contrib.auth.tokens.PasswordResetTokenGenerator"
-
-    def make_token(self, path):
-        return salted_hmac(
-            self.key_salt,
-            path,
-        ).hexdigest()[::2]
-
-    def check_token(self, token, path):
-        return constant_time_compare(
-            token,
-            salted_hmac(self.key_salt, path).hexdigest()[::2]
-        )
-
-path_token_generator = PathTokenGenerator()
+from cms.tokens import PreviewTokenGenerator
 
 
 class PublishedBase(models.Model):
@@ -41,6 +18,20 @@ class PublishedBase(models.Model):
     class Meta:
         abstract = True
 
+    def get_preview_url(self):
+        if not hasattr(self, 'get_absolute_url'):
+            return None
+
+        return f'{self.get_absolute_url()}?preview=1'
+
+    def get_public_preview_url(self):
+        if not hasattr(self, 'get_absolute_url'):
+            return None
+
+        token_generator = PreviewTokenGenerator(self.get_absolute_url())
+
+        return f'{self.get_absolute_url()}?preview={token_generator.make_token()}'
+
 
 class PublishedBaseSearchAdapter(SearchAdapter):
 
@@ -64,12 +55,6 @@ class OnlineBase(PublishedBase):
         ),
     )
 
-    def get_preview_url(self):
-        if not hasattr(self, 'get_absolute_url'):
-            return None
-
-        return f'{self.get_absolute_url()}?preview={path_token_generator.make_token(self.get_absolute_url())}'
-
     class Meta:
         abstract = True
 
